
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "initng.h"

#include <sys/types.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <alloca.h>
#include <assert.h>
#include <fnmatch.h>
#include <time.h>

#include "initng_active_db.h"
#include "initng_service_cache.h"
#include "initng_process_db.h"
#include "initng_global.h"
#include "initng_toolbox.h"
#include "initng_common.h"
#include "initng_static_data_id.h"
#include "initng_plugin_callers.h"

/* #####  ACTIVE_DB_FIND_BY   ######## */

/* return index of service in active data structure or -1 if not found */
active_h *active_db_find_by_exact_name(const char *service)
{
    active_h *current = NULL;

    assert(service);

    /* first, search for an exact match */
    while_active_db(current)
    {
        /* check if this service name is like service */
        if (strcmp(current->name, service) == 0)
            return (current);               /* return it */
    }

    /* did not find any */
    return NULL;
}

active_h *active_db_find_by_name(const char *service)
{
    assert(service);

    active_h *current = active_db_find_by_exact_name(service);

    if (current)
        return (current);

    /* do pattern matching only if service name does not contain wildcards */
    if (strchr(service, '*') || strchr(service, '?'))
        return (NULL);

    /* create service name mask for pattern matching */
    char *service_mask = initng_calloc(strlen(service) + 3, sizeof(char));

    strcpy(service_mask, "*/");
    strcat(service_mask, service);

    while_active_db(current)
    {

        /* then try to find alike name */

        /* if existing service is a pattern */
        if (strchr(current->name, '*') || strchr(current->name, '?'))
        {
            if (fnmatch(current->name, service, FNM_CASEFOLD) == 0)

            {
                D_("active_db_find_by_name(%s): Found\n", current->name);
                free(service_mask);
                return (current);
            }
        }
        else if (fnmatch(service_mask, current->name, FNM_CASEFOLD) == 0)
        {
            D_("active_db_find_by_name(%s): Found\n", current->name);
            free(service_mask);
            return (current);
        }
    }
    /* did not find any */
    free(service_mask);
    return NULL;
}

active_h *active_db_find_in_name(const char *service)
{
    active_h *current = NULL;

    assert(service);

    /* first, search for an exact match */
    while_active_db(current)
    {
        /* check if this service name is like service */
        if (strcmp(current->name, service) == 0)
            return (current);               /* return it */
    }

    /* then search for a word match */
    current = NULL;
    while_active_db(current)
    {
        /* check if this service name is like service */
        if (strstr(current->name, service))
        {
            D_("Warning, the return is not an exact match!\n");
            return (current);               /* return it */
        }
    }

    /* did not find any */
    return NULL;
}

/* return index of service in active data structure or -1 if not found */
active_h *active_db_find_by_service_h(service_h * service)
{
    active_h *current = NULL;

    assert(service);

    /* walk the active_db */
    while_active_db(current)
    {
        /* check if this service->from_service is like service */
        if (current->from_service && current->from_service == service)
            return (current);               /* return it */

    }

    return NULL;
}

/* returns pointer to active_h process belongs to, and sets process_type */
active_h *active_db_find_by_pid(pid_t pid)
{
    active_h *current = NULL;

    /* walk the active_db */
    while_active_db(current)
    {
        /* check for matching */
        if (current->start_process && current->start_process->pid == pid)
            return (current);
        if (current->stop_process && current->stop_process->pid == pid)
            return (current);
    }

    /* bad luck */
    return (NULL);
}





/* #########   STRUCT/CHAIN HANDLE  ######## */

/* add active to data structure */
/* remember to free service, if this fails */
int active_db_add(active_h * add_this)
{
    active_h *current = NULL;

    /* we have to get this data */
    assert(add_this);

    /* Look for duplicate */
    if ((current = active_db_find_by_name(add_this->name)))
    {
        /* TODO, shud add_this bee freed? */
        W_("active_db_add(%s): duplicate here\n", add_this->name);
        return (FALSE);
    }

    list_add(&add_this->list, &g.active_database.list);

    return (TRUE);
}

/* ############    ACTIVE_DB_ALLOC/FREE ##########  */

/* creates a new active_h at new_active */
active_h *active_db_new(const char *name)
{
    active_h *new_active = NULL;

    /* allocate a new active entry */
    new_active = (active_h *) initng_calloc(1, sizeof(active_h));       /* Allocate memory for a new active */

    /* out of memory? */
    if (!new_active)
    {
        F_("active_db_new(): Unable to allocate active, out of memory?\n");
        return (NULL);
    }

    new_active->name = i_strdup(name);

    INIT_LIST_HEAD(&(new_active->data.list));

    new_active->time_got_status = time(0);
    new_active->uptime = new_active->time_got_status;
    new_active->a_status = NONE;
    new_active->from_service = NULL;

    if (new_active->start_process)
    {
        process_db_free(new_active->start_process);
        new_active->start_process = NULL;
    }

    if (new_active->stop_process)
    {
        process_db_free(new_active->stop_process);
        new_active->stop_process = NULL;
    }
    return (new_active);
}

/* free some values in this one */
void active_db_free(active_h * pf)
{
    assert(pf);

    D_("(%s);\n", pf->name);

    /* look if there are plugins, that is intrested to now, this is freeing */
    mark_service(pf, FREEING);

    /* try free process structs, data and name */
    if (pf->start_process)
    {
        process_db_free(pf->start_process);
        pf->start_process = NULL;
    }
    if (pf->stop_process)
    {
        process_db_free(pf->stop_process);
        pf->stop_process = NULL;
    }

    /* remove every data entry */
    d_remove_all(&pf->data.list);

    if (pf->name)
        free(pf->name);

    free(pf);
}


/* clear database */
void active_db_free_all(void)
{
    active_h *current, *safe = NULL;

    while_active_db_safe(current, safe)
    {
        list_del(&current->list);
        active_db_free(current);
    }
}

/* #########  ACTIVE_DB_UTILS  #############  */

/* compensate time */
void active_db_compensate_time(time_t skew)
{
    active_h *current = NULL;

    /* walk the active_db */
    while_active_db(current)
    {
        /* change this time */
        current->time_got_status += skew;
	current->uptime += skew;
    }
}


/* update all service_h pointers in whole active_db */
void active_db_change_service_h(service_h * from, service_h * to)
{
    active_h *current = NULL;

    /* walk the active_db */
    while_active_db(current)
    {
        /* change this time */
        if (current->from_service == from)
            current->from_service = to;
    }
}


/* Walks thru db, and returns all services, that depends on this service */
active_h **active_db_dep_on_me(active_h * depend)
{
    char *string = NULL;
    active_h *current = NULL;
    active_h **to_ret = NULL;
    int i = 0;

    assert(depend);
    assert(depend->name);

    to_ret = (active_h **) initng_calloc(200, sizeof(active_h *));

    while_active_db(current)
    {
        /* we will hit ur self, sooner or later */
        if (current == depend)
            continue;
        /* dont return the current one */

        string = NULL;                      /* check if this service has any NEED flags */
        while ((string = active_db_get_next_string(&NEED, current, string)))
        {
            if (strcmp(depend->name, string) == 0)
            {
                D_("service %s depends (need) on %s!\n", current->name,
                   depend->name);
                to_ret[i] = current;
                i++;
            }
        }

        string = NULL;
        /* check if this service has any USE flags */
        while ((string = active_db_get_next_string(&USE, current, string)))
        {
            assert(string);
            if (strcmp(depend->name, string) == 0)
            {
                D_("service %s depends (use) on %s!\n", current->name,
                   depend->name);
                to_ret[i] = current;
                i++;
            }
        }

    }

    return (to_ret);
}

/* Returns a list of all deps, this active depends on */
active_h **active_db_next_service_this_depends_on(active_h * depend)
{
    char *string = NULL;
    active_h *tmp = NULL;
    active_h **to_ret = NULL;
    int i = 0;

    assert(depend);
    assert(depend->name);

    to_ret = (active_h **) initng_calloc(201, sizeof(active_h *));


    while ((string = active_db_get_next_string(&NEED, depend, string)))
    {
        if ((tmp = active_db_find_by_name(string)))
        {
            D_("service %s depends on me!\n", tmp->name);
            to_ret[i] = tmp;
            i++;
            /* make sure it wont grow to big */
            if (i >= 200)
                return (to_ret);
        }
    }

    string = NULL;
    /* check if this service has any USE flags */
    while ((string = active_db_get_next_string(&USE, depend, string)))
    {
        if ((tmp = active_db_find_by_name(string)))
        {
            D_("service %s depends on me!\n", tmp->name);
            to_ret[i] = tmp;
            i++;
            /* make sure it wont grow to big */
            if (i >= 200)
                return (to_ret);
        }
    }

    return (to_ret);
}


/* active_db_count counts a type, if null, count all */
int active_db_count(e_a_status a_status_to_count)
{
    int counter = 0;            /* actives counter */
    active_h *current = NULL;

    if (a_status_to_count == NONE)
    {
        /* ok, go thru all */
        while_active_db(current)
        {

            /* count almost all */
            switch (current->a_status)
            {
                case FAIL_STARTING:
                case FAIL_STOPPING:
                case START_DEP_FAILED:
                case STOP_DEP_FAILED:
                case STOPPED:
                    break;
                default:
                    counter++;
                    break;
            }
        }
    }
    else
    {
        /* ok, go thru all */
        while_active_db(current)
        {
            /* check if this is the status to count */
            if (current->a_status == a_status_to_count)
                counter++;
        }
    }

    /* return counter */
    return (counter);
}


/* calculate percent of processes started */
int active_db_percent_started(void)
{
    int starting = 0;
    int up = 0;
    int other = 0;
    float tmp = 0;

    active_h *current = NULL;

    /* walk the active_db */
    while_active_db(current)
    {
        switch (current->a_status)
        {
            case STARTING:
            case WAITING_FOR_START_DEP:
            case START_DEP_MET:
            case LOADING:
                /*D_("active_db_percent_started(): %s starting\n",
                   current->name); */
                starting++;
                break;
            case DONE:
            case RUNNING:
                /*D_("active_db_percent_started(): %s: up\n", current->name); */
                up++;
                break;
            default:
                other++;
                break;
        }
    }
    D_("active_db_percent_started(): up: %i   starting: %i  other: %i\n", up,
       starting, other);

    /* if no one starting */
    if (starting <= 0)
        return (100);

    if (up > 0)
    {
        tmp = 100 * (float) up / (float) (starting + up);
        D_("active_db_percent_started(): up/starting: %f percent: %i\n\n",
           (float) up / (float) starting, (int) tmp);
        return ((int) tmp);
    }
    return 0;
}

/* calculate percent of processes stopped */
int active_db_percent_stopped(void)
{
    int stopping = 0;
    int down = 0;
    int other = 0;
    float tmp = 0;
    active_h *current = NULL;

    while_active_db(current)
    {
        switch (current->a_status)
        {

            case STOP_LAUNCHING:
            case STOPPING:
            case WAITING_FOR_STOP_DEP:
            case STOP_DEP_MET:
            case STOP_MARKED:
	    case STOP_LAUNCHED:
                /*D_("active_db_percent_stopped(): %s stopping\n",
                   current->name); */
                stopping++;
                break;


            case STOPPED:
            case NONE:
                /*D_("active_db_percent_stopped(): %s: down\n", current->name); */
                down++;
                break;
    
	    default:
                other++;
                break;
        }
    }

    D_("active_db_percent_stopped(): down: %i   stopping: %i  other: %i\n",
       down, stopping, other);

    /* if no one stopping */
    if (stopping <= 0)
        return (100);

    if (down > 0)
    {
        tmp = 100 * (float) down / (float) (stopping + down);
        D_("active_db_percent_stopped(): down/stopping: %f percent: %i\n\n",
           (float) down / (float) stopping, (int) tmp);
        return ((int) tmp);
    }
    return 0;
}


/* ########     Data handlers  ############## */
char *active_db_get_string(s_entry * type, active_h * from_active)
{
    char *t = NULL;

    assert(from_active);

    /* look in active datas */
    if ((t = d_get_string(type, &from_active->data.list)))
        return (t);

    /* load a service file, if not exists */
    if (!from_active->from_service)
        get_service(from_active);

    /* look in service file data */
    if (from_active->from_service)
        return (service_db_get_string(type, from_active->from_service));

    return (NULL);
}

int active_db_get_int(s_entry * type, active_h * from_active)
{
    assert(from_active);

    if (d_is(type, &from_active->data.list))
        return (d_get_int(type, &from_active->data.list));

    if (!from_active->from_service)
        get_service(from_active);
    if (from_active->from_service)
        return (service_db_get_int(type, (from_active->from_service)));
    return (0);
}

char *active_db_get_next_string(s_entry * type, active_h * from_active,
                                char *last)
{
    char *r = NULL;

    assert(from_active);

    /* check active overlay data */
    if ((r = d_get_next_string(type, &from_active->data.list, last)))
        return (r);

    /* load service from disk, if nonexistant */
    if (!from_active->from_service)
        get_service(from_active);

    /* check service data */
    if (from_active->from_service)
        return (service_db_get_next_string
                (type, from_active->from_service, last));
    return (NULL);
}


int active_db_is(s_entry * type, active_h * from_active)
{
    assert(from_active);

    /* check active overlay data */
    if (d_is(type, &from_active->data.list))
        return (TRUE);

    /* load service from disk, if nonexistant */
    if (!from_active->from_service)
        get_service(from_active);

    /* check service data */
    if (from_active->from_service)
        if (service_db_is(type, from_active->from_service))
            return (TRUE);

    return (FALSE);
}

int active_db_count_type(s_entry * type, active_h * from_active)
{
    assert(from_active);
    if (!from_active->from_service)
        get_service(from_active);

    if (from_active->from_service)
        return (service_db_count_type(type, from_active->from_service));
    return (0);
}

