
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "initng.h"

#define _GNU_SOURCE
#include <stdlib.h>                         /* free() exit() */
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>

#include "initng_error.h"

#include "initng_global.h"
#include "initng_toolbox.h"
#include <assert.h>

const char *last_file = NULL;
const char *last_func = NULL;

int msgs = 0;
int lock_error_printing = 0;

#ifdef DEBUG
static void verbose_print(void)
{
    int i;

    if (lock_error_printing == 1)
        return;

    W_("This words will i look for in debug: ");
    for (i = 0; i < MAX_VERBOSES; i++)
        if (g.verbose_this[i])
            W_("  * \"%s\"\n", g.verbose_this[i]);
}

int verbose_add(const char *string)
{
    int i = 0;

    if (g.verbose == 1)
        g.verbose = 3;
    else
        g.verbose = 2;


    while (g.verbose_this[i] && i < MAX_VERBOSES)
        i++;

    if (i >= MAX_VERBOSES - 1)
    {
        F_("Can't add another \"%s\" debug trace\n", string);
        return (FALSE);
    }

    g.verbose_this[i] = i_strdup(string);

    verbose_print();

    if (g.verbose_this[i])
        return (TRUE);
    return (FALSE);
}

int verbose_del(const char *string)
{
    int i;
    int t = FALSE;

    for (i = 0; i < MAX_VERBOSES; i++)
        if (g.verbose_this[i] && strcasecmp(g.verbose_this[i], string) == 0)
        {
            free(g.verbose_this[i]);
            g.verbose_this[i] = NULL;
            t = TRUE;
        }
    verbose_print();
    return (t);
}
#endif


#ifdef DEBUG
void print_func(const char *file, const char *func)
{
    int i = 0;

    if (lock_error_printing == 1)
        return;
    lock_error_printing = 1;

    if (g.verbose == 2 || g.verbose == 3)
    {
        for (i = 0; i < MAX_VERBOSES; i++)
        {
            if (g.verbose_this[i])
            {
                if (g.verbose_this[i][0] == '%' && (g.verbose_this[i] + 1))
                {
                    if (strcasestr(file, (g.verbose_this[i]) + 1) ||
                        strcasestr(func, (g.verbose_this[i]) + 1))
                    {
                        lock_error_printing = 0;
                        return;
                    }
                }
                else
                {
                    if (strcasestr(file, g.verbose_this[i]) ||
                        strcasestr(func, g.verbose_this[i]))
                    {
                        i = 1;
                        break;
                    }
                }
            }
        }
    }

    if (g.verbose == 1 || i == 1)
    {
        if (last_file != file || last_func != func)
        {
            fprintf(stderr, "\n\n ** \"%s\", %s():\n", file, func);
        }

        last_file = file;
        last_func = func;
    }

    lock_error_printing = 0;
}
#endif

int print_error(e_mt mt, const char *file, const char *func, int line,
                const char *format, ...)
{
    int done;
    struct tm *ts;
    time_t t;


    assert(file);
    assert(func);

    if (lock_error_printing == 1)
        return (0);
    lock_error_printing = 1;


#ifdef DEBUG
    if (mt == MSG_VERBOSE)
    {
        int i;

        if (g.verbose == 1)
            goto yes;
        if (g.verbose == 2 || g.verbose == 3)
            for (i = 0; i < MAX_VERBOSES; i++)
                if (g.verbose_this[i])
                {
                    if (g.verbose_this[i][0] == '%'
                        && (g.verbose_this[i] + 1))
                    {
                        if (strcasestr(file, (g.verbose_this[i]) + 1) ||
                            strcasestr(func, (g.verbose_this[i]) + 1))
                        {
                            lock_error_printing = 0;
                            return (TRUE);
                        }
                    }
                    else
                    {
                        if (strcasestr(file, g.verbose_this[i]) ||
                            strcasestr(func, g.verbose_this[i]))
                        {
                            goto yes;
                        }
                    }
                }
        if (g.verbose == 3)
            goto yes;
        /* else */

        lock_error_printing = 0;
        return (TRUE);
    }
  yes:
#endif

    /* print the function name, if not set */
    if (last_file != file || last_func != func)
    {
        fprintf(stderr, "\n\n ** \"%s\", %s():\n", file, func);
    }

    last_file = file;
    last_func = func;

    if (g.i_am_init && getpid() != 1)
    {
        fprintf(stderr, " NON_INIT: ");
    }

    /* Dont fech time, until we know we wanna print on screen */
    t = time(0);
    ts = localtime(&t);


    switch (mt)
    {
        case MSG_FAIL:
            fprintf(stderr, " %.2i:%.2i:%.2i -- line:%i FAIL:\t", ts->tm_hour,
                    ts->tm_min, ts->tm_sec, line);
            break;
        case MSG_WARN:
            fprintf(stderr, " %.2i:%.2i:%.2i -- line:%i WARN:\t", ts->tm_hour,
                    ts->tm_min, ts->tm_sec, line);
            break;
        case MSG_VERBOSE:
            fprintf(stderr, " %.2i:%.2i:%.2i -- line:%i\t", ts->tm_hour,
                    ts->tm_min, ts->tm_sec, line);
            break;
    }

    msgs++;
    if (msgs > 20)
    {
        sleep(0.5);
        msgs = 0;
    }

    va_list arg;

    va_start(arg, format);

    done = vfprintf(stderr, format, arg);

    va_end(arg);

    lock_error_printing = 0;
    return (done);
}
